from minickt.gate import *

def writeOutCkt(ckt,file_name):
    inputs = []
    outputs = []
    wires = []
    gates = ckt.getPIGates()
    gates.extend(ckt.getGates()[ckt.po_end_index:])
    for gate in gates:
        if isinstance(gate,ConstZeroGate) or isinstance(gate,ConstOneGate):
            continue
        if gate.getFanInNum()==0:
            inputs.append(gate)
        elif gate.getFanOutNum()==0:
            outputs.append(gate)
        else:
            wires.append(gate)

    with open(file_name,'w') as writer:
        print >>writer,'//benchmark writen by minickt'
        io_name = outputs + inputs
        io_name = [gate.getName() for gate in io_name]
        ckt.circuit_name = ckt.circuit_name.partition('(')[0]
        print >>writer,'module {0} ({1});'.format(ckt.circuit_name,','.join(io_name))
        for o in outputs:
            print >>writer,'output {0};'.format(o.getName())
        for i in inputs:
            print >>writer,'input {0};'.format(i.getName())
        for w in wires:
            print >>writer,'wire {0};'.format(w.getName())
        for gate in ckt.getGates()[ckt.po_end_index:]:
            if isinstance(gate,ConstZeroGate) or isinstance(gate,ConstOneGate):
                continue
            gate_type = gate.getTypeStr()
            if '-' in gate_type:
                gate_type = gate_type.partition('-')[2]
            gate_type = gate_type.lower()
            gate_name = gate.getName()
            fanins = gate.getFanIns()
            gate_pins = []
            for fanin in fanins:
                if isinstance(fanin,ConstZeroGate):
                    gate_pins.append("1'b0")
                elif isinstance(fanin,ConstOneGate):
                    gate_pins.append("1'b1")
                else:
                    gate_pins.append(fanin.getName())
            print >>writer,'{0} ({1});'.format(gate_type,','.join([gate_name]+gate_pins))
        print >>writer,'endmodule'


# primitive-level parser but no buf

def isComment(line):
    if line[0:2]=='//':
        return True
    else:
        return False

def readStatement(reader):
    statement = ''
    while True:
        line = reader.readline()
        if not line:
            return False
        line = line.strip()
        if isComment(line):
            continue
        if line[-1] == ';' or line=='endmodule':
            return (statement+' '+line).strip()
        else:
            statement += ' '+line

def findFanins(gate,verilog_gate):
    if gate in verilog_gate['buf']:
        return verilog_gate['buf'][gate]
    return verilog_gate['buf'][buf_gate][0]

def clearBuf(verilog_gate):

    for keyword in verilog_gate:
        for gate in verilog_gate[keyword]:
            new_fanins = []
            for fi in verilog_gate[keyword][gate]:
                if fi in verilog_gate['buf']:
                    new_fanins.append(verilog_gate['buf'][fi][0])
                else:
                    new_fanins.append(fi)
            verilog_gate[keyword][gate] = new_fanins
    verilog_gate['buf'] = {}

def getCktInfo(file_name):

    ckt_info = {}

    const0 = False
    const1 = False
    verilog_module = ['module','endmodule']
    verilog_pin = {'input':[],'output':[],'wire':[]}
    verilog_gate = {'and':{},'or':{},'xor':{},'nor':{},'nand':{},'xnor':{},'buf':{},'not':{}}

    with open(file_name,'r') as reader:
        while True:
            statement = readStatement(reader)
            if not statement:
                break
            tokens = statement.strip(';').split()
            tokens = tokens if len(tokens)>1 or tokens[0]=='endmodule' else statement.strip(';').split('(')
            keyword = tokens[0]
            if keyword=='module':
                circuit_name = '{0}({1})'.format(tokens[1],file_name)
            elif keyword in verilog_pin:
                pin_names = [token.strip(',') for token in tokens[1:]]
                verilog_pin[keyword].extend(pin_names)
            elif keyword in verilog_gate:
                gate_pins = tokens[1].strip('()').split(',')
                gate_pins = [pin.strip() for pin in gate_pins]
                output = gate_pins[0]
                fanins = gate_pins[1:]
                fanins = ['const_zero' if fanin=="1'b0" else fanin for fanin in fanins]
                fanins = ['const_one' if fanin=="1'b1" else fanin for fanin in fanins]
                if 'const_zero' in fanins: const0 = True
                if 'const_one' in fanins: const1 = True
                verilog_gate[keyword][output] = fanins 
            else:
                pass

    clearBuf(verilog_gate)

    ckt_info['ckt_name'] = circuit_name
    ckt_info['output'] = verilog_pin['output']
    ckt_info['input'] = verilog_pin['input']
    ckt_info['wire'] = verilog_pin['wire']
    ckt_info['and'] = verilog_gate['and']
    ckt_info['or'] = verilog_gate['or']
    ckt_info['xor'] = verilog_gate['xor']
    ckt_info['nor'] = verilog_gate['nor']
    ckt_info['nand'] = verilog_gate['nand']
    ckt_info['xnor'] = verilog_gate['xnor']
    ckt_info['buf'] = verilog_gate['buf']
    ckt_info['not'] = verilog_gate['not']
    ckt_info['const0'] = const0
    ckt_info['const1'] = const1

    return ckt_info


from circuitManager import *
from equivalentCheck import *
from tempfile import NamedTemporaryFile
import os
import subprocess
import sys
import cmd
import time

class OriShell(cmd.Cmd,object):
    intro = \
"""
================================
           miniCkt

     * author: dokelung
     # use "?" or "cmd"         
================================
"""
    prompt = '\033[94m[ miniCkt ] >> \033[0m'
    #prompt = '[ miniCkt ] >> '

    def __init__(self):
        super(Shell, self).__init__()
        self.cktmgr = CircuitManager()

    def do_read(self,arg):
        """
        read circuit file by specified parser and construct a circuit
        SYNOPSIS: read <circuit file> by <parser>
        DESCRIPTION:
            1.  each parser is a python module located in directory "cktParser"
                and it must have a global function "getCktInfo(file_name)"
                which return a circuit infomation dictionary
                to know the details of this dictionary,
                user can reference to "primitive.py" in "cktParser"
                note that file_name is any kind of circuit file, e.g. verilog
            2.  also, "cktParser" is a python package
            3.  command "ls_parser" can list all avalible parsers
        """
        args = self.parseArg(arg)
        try:
            file_name = args[0]
            assert args[1]=='by'
            parser= args[2]

            exec 'from cktParser import {0}'.format(parser)
            exec 'ckt_info = {0}.getCktInfo("{1}")'.format(parser,file_name)

            print 'read {0} ...'.format(file_name)
            ckt = Circuit()
            ckt.constructCkt(ckt_info)
            self.cktmgr.addCkt(ckt)
            self.cktmgr.transToCkt(len(self.cktmgr.getCkts())-1)
            print 'success:', 'circuit "{0}" is in circuit manager now'.format(ckt)
        except:
            print 'error: reading failed'

    def do_write(self,arg):
        """
        write current circuit out to a verilog of primitive-gate type
        SYNOPSIS: write <file name> by <writer>
        DESCRIPTION:
            1.  if user want to modify the module's name, please use cmd "chname"
                to change the name of circuit(module) first
                note that the name with '()' may result in error!
            2.  each writer is a python module located in directory "cktWriter"
                and it must have a global function "writeOutCkt(ckt,file_name)"
                for writing out circuit
                user can reference to "primitive.py" in "cktWriter"
                note that file_name is any kind of circuit file, e.g. verilog
            2.  also, "cktWriter" is a python package
            3.  command "ls_writer" can list all avalible writers
        """
        #try:
        args = self.parseArg(arg)
        ckt = self.cktmgr.getCurrentCkt()

        file_name = args[0]
        assert args[1]=='by'
        writer = args[2]

        exec 'from cktWriter import {0}'.format(writer)
        exec '{0}.writeOutCkt(ckt,"{1}")'.format(writer,file_name)

        print 'success:', 'circuit "{0}" has been writen out'.format(ckt)

        #except:
        #    print 'error: cannot write out circuit'

    def do_cmd(self,arg):
        """
        list all cmds and their brief descriptions
        """
        try:
            print '=================================='
            print '  command list'
            print '=================================='
            commands = [name for name in dir(self) if name[0:3]=='do_']
            for command in commands:
                doc = eval('self.'+command+'.__doc__')
                if not doc==None:
                    doc = doc.strip()
                    doc = doc.split('\n')
                    print '{cmd:12}: {state}'.format(cmd=command[3:], state=doc[0])
        except:
            print 'error: cannot list commands'

    def do_ls(self,arg):
        """
        list items
        SYNOPSIS: ls <option>
        OPTION:
            -c = (cmd)list_ckt: list circuits in the circuit manager
            -p = (cmd)ls_parser: list avalible parsers
            -w = (cmd)ls_writer: list avalible writers
        """
        try:
            args = self.parseArg(arg)
            if args[0] == 'c':
                self.excCmd('ls_ckt')
            elif args[0] == 'p':
                self.excCmd('ls_parser')
            elif args[0] == 'w':
                self.excCmd('ls_writer')
            else:
                print 'error: no this option'
        except:
            print 'error: cannot list, use "?ls" to check the usage'

    def do_ls_parser(self,arg):
        """
        list all avalible parsers in package(directory) cktParser
        """
        try:
            files = os.listdir('cktParser')
            files = [file for file in files if '.py' in file and '__init__' not in file and '.pyc' not in file]
            files = [file.split('.')[0] for file in files]
            print 'there are {0} parsers in cktParser'.format(len(files))
            for index, file in enumerate(files):
                print '[{0}]'.format(index), file
        except:
            print 'error: cannot list parsers'

    def do_ls_writer(self,arg):
        """
        list all avalible writers in package(directory) cktWriter
        """
        try:
            files = os.listdir('cktWriter')
            files = [file for file in files if '.py' in file and '__init__' not in file and '.pyc' not in file]
            files = [file.split('.')[0] for file in files]
            print 'there are {0} parsers in cktWriter'.format(len(files))
            for index, file in enumerate(files):
                print '[{0}]'.format(index), file
        except:
            print 'error: cannot list writers'

    def do_sec(self,arg):
        """
        check structural equivalence
        SYNOPSIS: sec <[ckt id.]gate name> <[ckt id.]gate name>
        DESCRIPTION:
            1. if just give gate name, miniCkt take it as a gate in current ckt
            2. user can use "ls -c" or "ls_ckt" to get "ckt id"
        EXAMPLE:
            ex1. [ miniCkt ] >> sec 0.out 2.out
                 ---> sec for two circuit
            ex2. [ miniCkt ] >> sec n4 n5
                 ---> sec for two gate cones
        """
        try:
            args = self.parseArg(arg)
            if '.' in args[0]:
                ckt1_id, gate_name1 = args[0].split('.')
                ckt1_id = int(ckt1_id)
            else:
                ckt1_id = self.cktmgr.getToken()
                gate_name1 = args[0]

            if '.' in args[1]:
                ckt2_id, gate_name2 = args[1].split('.')
                ckt2_id = int(ckt2_id)
            else:
                ckt2_id = self.cktmgr.getToken()
                gate_name2 = args[1]

            ckt1 = self.cktmgr.getCkt(ckt1_id)
            ckt2 = self.cktmgr.getCkt(ckt2_id)
            gate1 = ckt1.getGateFromName(gate_name1)
            gate2 = ckt2.getGateFromName(gate_name2)

            if checkStructuralEquivalence(gate1,gate2):
                print 'structural eq'
            else:
                print 'structural uneq'
        except:
            print 'error: checking error'

    def do_fec(self,arg):
        """
        check functional equivalence
        SYNOPSIS: sec <ckt id> <ckt id>
        DESCRIPTION:
            1.  fec checks functional equivalence with miter
                and now just supports "single output" curcuit.
                So make sure two things:
                    1. The specified ckts should be single output ckt.
                    2. Their inputs can be mapped.
                       (It means the corresponding inputs should have same name.)
            2.  If the miter's satisfiability is SAT, these two ckts are not 
                functional equivalence, otherwise the are functional equivalence.
            3.  It use lingeling SAT solver and TST encoding to solve
                miter's circuit SAT.
            4.  user can use "ls -c" or "ls_ckt" to get "ckt id"
        """
        try:
            args = self.parseArg(arg)
            ckt1_id = int(args[0])
            ckt2_id = int(args[1])
            ckt1 = self.cktmgr.getCkt(ckt1_id)
            ckt2 = self.cktmgr.getCkt(ckt2_id)

            for g in ckt1.getGatesOnce(): 
                if g not in ckt1.getPIGates():
                    g.nameTranslate('_m1')
            for g in ckt2.getGatesOnce(): 
                if g not in ckt2.getPIGates():
                    g.nameTranslate('_m2')

            from cktWriter import primitive
            primitive.writeOutCkt(ckt1,'c1')
            primitive.writeOutCkt(ckt2,'c2')

            inputs = []
            outputs = []
            wires = []
            gates = []

            for file_name in ['c1','c2']:
                with open(file_name,'r') as reader:
                    for line in reader:
                        line = line.strip()
                        if line[0:2]=='//':
                            continue
                        else:
                            items = line.split()
                            keyword = items[0]
                            if keyword in ['module','endmodule']:
                                continue
                            if keyword == 'input':
                                inputs.append(items[1].strip(';'))
                            elif keyword == 'output':
                                outputs.append(items[1].strip(';'))
                            elif keyword == 'wire':
                                wires.append(items[1].strip(';'))
                            else:
                                gates.append(line)

            inputs = set(inputs)

            io_name = ['out'] + list(inputs)

            with open('miter','w') as writer:
                print >>writer,'//benchmark writen by minickt'
                print >>writer,'module {0} ({1});'.format('miter',','.join(io_name))
                for i in inputs:
                    print >>writer,'input {0};'.format(i)
                print >>writer, 'output out;'
                for w in wires+outputs:
                    print >>writer,'wire {0};'.format(w)
                for g in gates:
                    print >>writer, g
                print >>writer,'xor (out,{0},{1});'.format(outputs[0],outputs[1])
                print >>writer,'endmodule'

            os.system('rm c1')
            os.system('rm c2')

            self.excCmd('read miter by primitive')
            self.excCmd('chname miter')
            self.excCmd('!rm miter')
            self.excCmd('sat by lingeling -f with TST')

        except:
            print 'error: checking error'

    def do_reset(self,arg):
        """
        reset the value of all gates
        """
        try:
            ckt = self.cktmgr.getCurrentCkt()
            ckt.resetAllGates()
        except:
            print 'error: cannot reset'

    def do_reason(self,arg):
        """
        reason the ckt under current assignment
        DESCRIPTION:
            1. user can use cmd "set" to assign gate value
            2. use cna use cmd "reset" to reset all gate values
            3. also, cmd "sg" with option "-v" can check the
               result of reasoning
        """
        try:
            ckt = self.cktmgr.getCurrentCkt()
            t1 = time.time()
            if ckt.reason():
                print 'reason success'
            else:
                print 'conflict'
            t2 = time.time()
            print '* reason time: {time} sec'.format(time=t2-t1)
        except:
            print 'error: cannot reason'

    def do_sat_reason(self,arg):
        """
        use sat solving for totally reasoning
        SYNOPSIS: sat_reason <gate types/gate names> <value>
        GATE_TYPE:
            pi, po, and, or, buf, not, nand, nor, xor, xnor, all
        VALUE:
            logical one  -> "1" or "t"
            logical zero -> "0" or "f"
        DESCRIPTION:
            use tseitin transformation to encode
            use lingeling and its default setting to solve
        """
        try:
            self.excCmd('encode by PTST to tempf')
            self.excCmd('reset')
            self.excCmd('set {0}'.format(arg))
            ckt = self.cktmgr.getCurrentCkt()
            with open('tempf','a') as f:
                for gate in ckt.getGates()[ckt.po_end_index:]+ckt.getPIGates():
                    if not gate.getValue()==None:
                        if gate.getValue()==True:
                            print >>f,gate.getId(True),
                            print >>f,'0'
                        elif gate.getValue()==False:
                            print >>f,-gate.getId(True),
                            print >>f,'0'
            os.system('./bin/lingeling -f tempf')
            os.system('rm tempf')
        except:
            print 'error: cannot do sat reasoning'

    def do_encode(self,arg):
        """
        encode circuit to a cnf file
        SYNOPSIS: encode by <method> to <file>
        METHOD:
            PTST:   pure tseitin transformation(no assign output) 
            TST:    tseitin transformation
            NTST:   new tseitin transformation(includes additional constraints)
            TST-F:  TST but assign output=False
            NTST-F: NTST but assign output=False
            RTST:   TST with reasoning
        DESCRIPTION:
            this cmd does not generate header of cnf file
        """
        try:
            ckt = self.cktmgr.getCurrentCkt()
            args = self.parseArg(arg)
            assert args[0]=='by'
            assert args[2]=='to'

            file_name = args.pop()
            
            #print 'id-table'
            #for gate in ckt.getGatesInTopologicalOrder():
            #    print gate,gate.getId(True)
            #with open('test1.tb','w') as writer:
            #    for gate in ckt.getGates():
            #        print >>writer, gate.getId(), gate

            constraint_type = {}

            t1 = time.time()
            if args[1]=='PTST':
                ckt.encTseitin(file_name,option='pure')
            elif args[1]=='TST' or args[1]=='TST-F':
                if args[1]=='TST':
                    ckt.encTseitin(file_name)
                elif args[1]=='TST-F':
                    ckt.encTseitin(file_name,option='unsat')
            elif args[1]=='NTST' or args[1]=='NTST-F':
                if args[1]=='NTST':
                    ckt.encTseitin(file_name)
                elif args[1]=='NTST-F':
                    ckt.encTseitin(file_name,option='unsat')
                with open(file_name,'a') as writer:
                    for gate in ckt.getPIGates()+ckt.getGates()[ckt.po_end_index:]:
                        added_clas, types = gate.getEncodeConstraint()
                        for type in types:
                            if constraint_type.has_key(type):
                                constraint_type[type] += types[type]
                            else:
                                constraint_type[type] = types[type]
                        for cla in added_clas:
                            for lit in cla:
                                print >>writer,lit,
                            print >>writer,'0'
            elif args[1]=='RTST':
                ckt.encReason(file_name)
            t2 = time.time()
            print '--------<other constraint>--------'
            for type in constraint_type:
                print 'add {0} constraint X {1}'.format(type,constraint_type[type])
            print '----------------------------------'
            print '\n* encode time: {time} sec \n'.format(time=t2-t1)
        except:
            print 'error: cannot encode'

    def do_sat(self,arg):
        """
        solve circuit sat by specified sat solver
        SYNOPSIS: sat by <sat solver & option> with <encoding method>
        DESCRIPTION:
            1. the excutable file of sat solver must be compiled
               and placed in the directory "bin"
            2. if user want to specify the solving setting, the setting must
               be given in "sat solver & option" with 
               example:
                    sat by lingeling -f with tseitin
            3. user can add new solver by modifying the function "do_sat"
               in "core/cmdManager"
            4. the supported encoding method can reference to cmd "encode"
               use "?encode" to know the details
        """
        try:
            args = arg.partition('by')[2]
            solver_setting, _ ,encoding_method = args.partition('with')
            solver_setting = solver_setting.strip()
            encoding_method = encoding_method.strip()
            with NamedTemporaryFile('w+t') as f:
                self.excCmd('encode by {method} to {fname}'.format(method=encoding_method,fname=f.name))
                os.system('./bin/{sat} {fname}'.format(sat=solver_setting,fname=f.name))

            #self.excCmd('encode')
        except:
            print 'error: cannot solve'

    def do_ls_ckt(self,arg):
        """
        list all (current) circuits in the circuit manager
        """
        try:
            print 'there are {0} circuits now'.format(len(self.cktmgr.getCkts()))
            for index, ckt in enumerate(self.cktmgr.getCkts()):
                if index==self.cktmgr.getToken():
                    print '[{0}]'.format(index), ckt, '<--- current'
                else:
                    print '[{0}]'.format(index), ckt
        except:
            print 'error: cannot list circuits'

    def do_eval(self,arg):
        """
        evaluate each gate in circuit
        """
        try:
            ckt = self.cktmgr.getCurrentCkt()
            t1 = time.time()
            ckt.evaluateCkt()
            t2 = time.time()
            print '* eval time: {time} sec'.format(time=t2-t1)
        except:
            print 'error: cannot evaluate'

    def do_cc(self,arg):
        """
        change current circuit to i-th circuit of circuit manager
        SYNOPSIS: cc <i>
        """
        try:
            self.cktmgr.transToCkt(int(arg))
        except:
            print 'error: no this circuit'

    def do_rm(self,arg):
        """
        remove circuit of circuit manager
        SYNOPSIS: rm [i]
        DESCRIPTION:
            if option "i" is given, remove i-th circuit of circuit manager
            otherwise, remove the current circuit
        """
        try:
            if not arg:
                self.cktmgr.rmCktByIndex(self.cktmgr.getToken())
            else:
                self.cktmgr.rmCktByIndex(int(arg))
        except:
            print 'error: cannot remove this circuit'

    def do_chname(self,arg):
        """
        change name of current circuit (set a new name for it)
        SYNOPSIS: chname <new name>
        """
        try:
            self.cktmgr.changeCktName(arg.strip())
        except:
            print 'error: cannnot change circuit name'

    def do_loadsc(self,arg):
        """
        load a script file and excute the commands in it
        SYNOPSIS: loadsc <script file> 
        """
        with open(arg,'r') as reader:
            for line in reader:
                line = line.strip()
                print '-------------------------'
                print 'EXC({0})'.format(line)
                print '-------------------------'
                self.excCmd(line)

    def do_shell(self,arg):
        """
        excute the command as in shell
        SYNOPSIS: shell <command>
        DESCRIPTION:
            user can simply use symbol ! to run cmd in shell
        EXAMPLE:
            to list file names of current directory in long format
            [ miniCkt ] >> shell ls -l
            [ miniCkt ] >> ! ls -l
        """
        os.system(arg)

    def do_show(self,arg):
        """
        show graph of circuit
        SYNOPSIS: show <option>
        OPTION:
            -c = (cmd)show_ckt: show a circuit graph
            -r = (cmd)show_ratio: show a pie chart of gates ratio
        """
        try:
            args = self.parseArg(arg)
            if args[0] == 'c':
                self.excCmd('show_ckt')
            elif args[0] == 'r':
                self.excCmd('show_ratio')
            else:
                print 'error: no this option'
        except:
            print 'error: cannot show, use "?show" to check the usage'

    def do_show_ckt(self,arg):
        """
        show the circuit graph
        """
        try:
            self.cktmgr.getCurrentCkt().showCircuitGraph()
        except:
            print 'error: cannot show the graph'

    def do_show_ratio(self,arg):
        """
        show the pie chart for listing gates ratio
        """
        try:
            self.cktmgr.getCurrentCkt().showGatesRatio()
        except:
            print 'error: cannot show the ratio of gates'

    def do_get(self,arg):
        """
        get the properties of circuit
        SYNOPSIS: get <circuit property> [>,>>log file]
        DESCRIPTION:
            user can output the infomation to specified file by notations
            ">"(write to log file) or ">>"(append to log file)
        CIRCUIT PROPERTY:
            ckt_name:       show the current circuit name
            avg_fanout_num: show the average number of fanout
            max_fanout_num: show the max number of fanout
            var_fanout_num: show the variance of number of fanout
            avg_fanin_num:  show the average number of fanin
            max_fanin_num:  show the max number of fanin
            var_fanin_num:  show the variance of number of fanin
            gate_num:       list the gate numbers and ratio
            path_num:       list the path numbers of circuit
            depth:          show the depths of all PO gates
            probability:    list the probabilities of all PO gates
                            also the cone gates whose probability=1 or =0
            consistency:    show the consistent rate
                            consistent => all fanouts' phases are the same
                            consistency = #consistent gate/#gate
            sop:            show the sum of product expression of POs 
        """
        try:
            ckt = self.cktmgr.getCurrentCkt()
            args = self.parseArg(arg)
            outplace = sys.stdout
            file_open = False
            if '>' in args[-1] or '>>' in args[-1]:
                try:
                    out_info = args[-1]
                    if '>>' in out_info:
                        file_name = out_info.rpartition('>')[2]
                        outplace = open(file_name,'a')
                    elif '>' in out_info:
                        file_name = out_info.rpartition('>')[2]
                        outplace = open(file_name,'w')
                    file_open = True
                except:
                    print 'error: please use "?get" to see the usage'
                    return False
                args = args[:-1]
            if 'probability' in args:
                for po in ckt.getPOGates():
                    print >>outplace,'{0} ==> probability={1}'.format(po,po.getProbability())
                    print '------------------------------------'
                    ckt.traverseCone(po)
                    for gate in ckt.getGates():
                        if gate.isFlagUp():
                            prob = gate.getProbability()
                            if prob <= 0:
                                print >>outplace,'{0} ==> probability={1}'.format(gate,prob)
                            elif prob >= 1:
                                print >>outplace,prob,'{0} ==> probability={1}'.format(gate,prob)
            if 'avg_fanout_num' in args:
                print >>outplace,'average #fanout = {0}'.format(ckt.getAvgFanOutNum())
            if 'max_fanout_num' in args:
                print >>outplace,'max #fanout = {0}'.format(ckt.getMaxFanOutNum())
            if 'var_fanout_num' in args:
                print >>outplace,'variance of #fanout = {0}'.format(ckt.getFanOutNumVariance())
            if 'avg_fanin_num' in args:
                print >>outplace,'average #fanin = {0}'.format(ckt.getAvgFanInNum())
            if 'max_fanin_num' in args:
                print >>outplace,'max #fanin = {0}'.format(ckt.getMaxFanInNum())
            if 'var_fanin_num' in args:
                print >>outplace,'variance of #fanin = {0}'.format(ckt.getFanInNumVariance())
            if 'depth' in args:
                for po in ckt.getPOGates():
                    print >>outplace,'{0} ==> depth={1}'.format(po,po.getDepth())
            if 'path_num' in args:
                for po in ckt.getPOGates():
                    print >>outplace,'{0} ==> path_num={1}'.format(po,po.getPathNum())
            if 'ckt_name' in args:
                print >>outplace,'circuit name = {0}'.format(ckt)
            if 'gate_num' in args:
                gate_count = ckt.getGateCount()
                labels = ['PI', 'AND', 'OR', 'XOR', 'XNOR', 'NAND', 'NOR', 'NOT', 'BUF', 'Const0', 'Const1']
                sizes = []
                sizes.append(ckt.getPiNum())
                sizes.append(ckt.getAndNum())
                sizes.append(ckt.getOrNum())
                sizes.append(ckt.getXorNum())
                sizes.append(ckt.getXnorNum())
                sizes.append(ckt.getNandNum())
                sizes.append(ckt.getNorNum())
                sizes.append(ckt.getNotNum())
                sizes.append(ckt.getBufNum())
                sizes.append(ckt.getConstZeroNum())
                sizes.append(ckt.getConstOneNum())
                print >>outplace,'total gate count = {0}'.format(gate_count)
                print '------------------------------------'
                for index, label in enumerate(labels):
                    print >>outplace,'#{0} = {1}({2})'.format(label,sizes[index],float(sizes[index])/float(gate_count))
            if 'consistency' in args:
                print >>outplace,'consistent rate = {0}'.format(ckt.getPhaseConsistentRate())
            if 'sop' in args:
                for po in ckt.getPOGates():
                    print >>outplace,'{0} ==> sop =\n{1}'.format(po,po.getSOP())
            if file_open:
                outplace.close()
        except:
            print 'error: cannot get'
            if file_open:
                outplace.close()

    def do_set(self,arg):
        """
        set value of gate in circuit
        SYNOPSIS: set <gate types/gate names> <value>
        GATE_TYPE:
            pi, po, and, or, buf, not, nand, nor, xor, xnor, all
        VALUE:
            logical one  -> "1" or "t"
            logical zero -> "0" or "f"
        DESCRIPTION:
            user can use script file to set
            just write set commands in script file
            and excute command loadsc script
        EXAMPLE:
            1. set directlly
                [ miniCkt ] >> set pi n01 n02 t
            2. set by script file
                [ miniCkt ] >> loadsc setting_script
                ---------------------------
                contents of setting_script
                ---------------------------
                set pi t
                set n01 n02 t
                --------------------------- 
            note that n01, n02 are the name of gates
        """
        try:
            ckt = self.cktmgr.getCurrentCkt()
            args = self.parseArg(arg)
            value = args.pop()
            if value=='1' or value=='t':
                value = True
            elif value=='0' or value=='f':
                value = False
            args = [arg.strip(',') for arg in args ]
            gates = self.collectGates(ckt,args)
            for gate in gates:
                gate.setValue(value)
        except:
            print 'error: cannot set'
        
    def do_sg(self,arg):
        """
        see the gates of current circuit
        SYNOPSIS: sg <gate types/gate names> [option] [k]
        GATE_TYPE:
            pi, po, and, or, buf, not, nand, nor, xor, xnor, const0, const1, all
        OPTION:
            -v: also see the value of gates
            -i: also see the fanins
            -o: also see the fanouts
        DESCRIPTION:
            1. k is an integer that indicates how many gates listed in a line
               default value: k=5, note that if user use option "-i" or "-o",
               k will be set 1(k=1) automatically
            2. user also can use gate type "all" to see all gates in circuit
        EXAMPLE:
            [ miniCkt ] >> sg n01 n02 and or -iov 3
            note that n01, n02 are the gate names
        """
        try:
            ckt = self.cktmgr.getCurrentCkt()
            #args = self.parseArg(arg)
            args = arg.strip().split()

            k = 5

            try:
                k = int(args[-1])
                args.pop()
            except:
                pass

            option = ''
            #if not args[-1] in ['pi', 'po', 'and', 'or', 'buf', 'not', 'nand', 'nor', 'xor', 'xnor', 'const0', 'const1', 'all']:
            #    option = args[-1]
            #    if 'i' in option or 'o' in option:
            #        k=1
            if args[-1][0]=='-':
                option = args.pop()

            args = [arg.strip(',') for arg in args]
            gates = self.collectGates(ckt,args)

            if 'i' in option or 'o' in option:
                k=1

            for index, gate in enumerate(gates):
                if 'v' in option:
                    print '{0}={1}'.format(gate,gate.getValue()),
                else:
                    print gate,

                if 'i' in option:
                    print '\n    -fanin-> {0}'.format(gate.getFanIns()),
                if 'o' in option:
                    print '\n    -fanout-> {0}'.format(gate.getFanOuts()),
                if index%k==k-1:
                    print '' 
            if not len(gates)%k==0:
                print ''         
        except:
            print 'error: cannot see the gates'

    def do_gen_cone_ckt(self,arg):
        """
        genearate a independent circuit by getting cone of specifed gates
        SYNOPSIS: gen_cone_ckt <gate name> [cone ckt name]
        DESCRIPTION:
            it is a combinational command
            gen_cone_ckt <gate name> [cone ckt name]:
                1. cone <gate name> [cone ckt name]
                2. write by primitive
                3. rm
                4. read by primitive
        """
        try:
            with NamedTemporaryFile('w+t') as f:
                self.excCmd('cone {args}'.format(args=arg))
                self.excCmd('write {fname} by primitive'.format(fname=f.name))
                self.excCmd('rm')
                self.excCmd('read {fname} by primitive'.format(fname=f.name))
        except:
            print 'error: cannot generate cone ckt'

    def do_cone(self,arg):
        """
        generate a sub-circuit by getting cone of specified gate
        SYNOPSIS: cone <gate name> [cone ckt name]
        DESCRIPTION:
            this cmd will generate a cone circuit by given name of gate
            also, this circuit will be added to circuit manager as a new circuit
            its name can be specified by "cone ckt name"
            if user wants to let cone circuit be an independent circuit, please
            write it out and read it again
            or can use combinational cmd "gen_cone_ckt" in convenience
        """
        try:
            ckt_name = 'coneckt'
            args = self.parseArg(arg)
            if len(args)==2:
                gate_name = args[0]
                ckt_name = args[1]
            else:
                gate_name = args[0]
            ckt = self.cktmgr.getCurrentCkt()
            cone = ckt.getConeCircuit(ckt.getGateFromName(gate_name))
            cone.setName(ckt_name)
            self.cktmgr.addCkt(cone)
            self.cktmgr.transToCkt(len(self.cktmgr.getCkts())-1)
            print 'success: generate cone of {0}'.format(cone.getGateFromName(gate_name))
        except:
            print 'error: cannot generate cone circuit'

    def do_exit(self,arg):
        """
        terminate this tool
        """
        exit()

    def excCmd(self,cmd):
        l = self.precmd(cmd)
        r = self.onecmd(l)
        r = self.postcmd(r,l)

    def parseArg(self,arg):
        args = arg.strip().split()
        args = [arg.strip('-') for arg in args]
        return args

    def collectGates(self,ckt,args):

        gates = []

        if 'pi' in args:
            gates.extend(ckt.getPIGates())
        if 'po' in args:
            gates.extend(ckt.getPOGates())
        if 'and' in args:
            gates.extend(ckt.getAndGates())
        if 'or' in args:
            gates.extend(ckt.getOrGates())
        if 'buf' in args:
            gates.extend(ckt.getBufGates())
        if 'not' in args:
            gates.extend(ckt.getNotGates())
        if 'nand' in args:
            gates.extend(ckt.getNandGates())
        if 'nor' in args:
            gates.extend(ckt.getNorGates())
        if 'xor' in args:
            gates.extend(ckt.getXorGates())
        if 'xnor' in args:
            gates.extend(ckt.getXnorGates())
        if 'const0' in args:
            gates.extend(ckt.getConstZeroGates())
        if 'const1' in args:
            gates.extend(ckt.getConstOneGates())
        if 'all' in args:
            gates.extend(ckt.getGates())

        for gate_name in args:
            try:
                gates.append(ckt.getGateFromName(gate_name))
            except:
                pass

        return list(set(gates))

class Shell(OriShell):

    def __init__(self):
        super(OriShell, self).__init__()
        self.cktmgr = CircuitManager()

    def do_collect(self,arg):
        """
        collect circuit info and print
        SYNOPSIS: collect [>,>>log file]
        DESCRIPTION:
            log file is a cvs file
        """
        try:
            outplace = sys.stdout
            args = self.parseArg(arg)
            title = False
            file_open = False
            if not len(args)==0:
                if '>' in args[-1] or '>>' in args[-1]:
                    try:
                        out_info = args[-1]
                        if '>>' in out_info:
                            file_name = out_info.rpartition('>')[2]
                            outplace = open(file_name,'a')
                        elif '>' in out_info:
                            file_name = out_info.rpartition('>')[2]
                            outplace = open(file_name,'w')
                            title = True
                        file_open = True
                    except:
                        print 'error: please use "?collect" to see the usage'
                        return False

            ckt = self.cktmgr.getCurrentCkt()
            ckt_info = []
            ckt_info.append( ('pi_num',ckt.getPiNum()) )
            ckt_info.append( ('and_num',ckt.getAndNum()) )
            ckt_info.append( ('or_num', ckt.getOrNum()) )
            ckt_info.append( ('xor_num', ckt.getXorNum()) )
            ckt_info.append( ('nand_num', ckt.getNandNum()) )
            ckt_info.append( ('nor_num', ckt.getNorNum()) )
            ckt_info.append( ('xnor_num',ckt.getXnorNum()) )
            ckt_info.append( ('buf_num',ckt.getBufNum()) )
            ckt_info.append( ('not_num',ckt.getNotNum()) )

            for i in range(len(ckt_info)):
                ckt_info.append( (ckt_info[i][0]+'_rate',float(ckt_info[i][1])/float(ckt.getGateCount())) )
        
        
            ckt_info.append( ('gate_count',ckt.getGateCount()) )
            ckt_info.append( ('probability',ckt.getPOProbability(0)) )
            ckt_info.append( ('avg_fanout_num',ckt.getAvgFanOutNum()) )
            ckt_info.append( ('max_fanout_num',ckt.getMaxFanOutNum()) )
            ckt_info.append( ('fanout_num_variance',ckt.getFanOutNumVariance()) )
            ckt_info.append( ('avg_fanin_num',ckt.getAvgFanInNum()) )
            ckt_info.append( ('max_fanin_num',ckt.getMaxFanInNum()) )
            ckt_info.append( ('fanin_num_variance',ckt.getFanInNumVariance()) )
            ckt_info.append( ('depth',ckt.getPOGates()[0].getDepth()) )
            ckt_info.append( ('path_num',ckt.getPOGates()[0].getPathNum()) )
        
            ckt_info = [(info[0],str(info[1])) for info in ckt_info]
            ckt_info_name = [info[0] for info in ckt_info]
            ckt_info_property = [str(info[1]) for info in ckt_info]

            if title:
                print >>outplace, ','.join(ckt_info_name)
            print >>outplace,','.join(ckt_info_property)
        
            if file_open:
                outplace.close()

        except:
            print 'error: cannot collect ckt info'
            if file_open:
                outplace.close()
            return False


from sop import *
import copy
import itertools

class Gate(object):
    
    def __init__(self,id):
        super(Gate, self).__init__()
        self.id = id
        self.fanins = []
        self.fanouts = []
        self.flag = False
        self.ispo = False
        self.value = None
        self.probability = None
        self.depth = None
        self.path_num = None
        self.SOP = None
        self.structure_equal_gates = []
        return self

    def __str__(self):
        return '{0}:{1}'.format(self.type_str,self.getName())

    def __repr__(self):
        return '{0}:{1}'.format(self.type_str,self.getName())

    def getDepth(self):
        if not self.depth==None:
            return self.depth
        else:
            self.depth = max([fanin.getDepth() for fanin in self.fanins])+1
            return self.depth

    def getPathNum(self):
        if not self.path_num==None:
            return self.path_num
        else:
            self.path_num = sum([fanin.getPathNum() for fanin in self.fanins])
            return self.path_num

    def getId(self,increase=False):
        if increase:
            return self.id+1
        return self.id

    def getName(self):
        return self.name

    def getFanOuts(self):
        return self.fanouts

    def getFanOutNum(self):
        return len(self.fanouts)

    def getFanIns(self):
        return self.fanins

    def getFanInNum(self):
        return len(self.fanins)

    def getFanOut(self,index):
        return self.fanouts[index]

    def getFanIn(self,index):
        return self.fanins[index]

    def getTypeStr(self):
        return self.type_str

    def getValue(self):
        return self.value

    def setValue(self,value):
        self.value = value

    def deduceValue(self,value):
        if not self.value == None and not self.value==value:
            return False
        else:
            self.value = value
            return True

    def setName(self,name):
        self.name = name
        return self.id

    def addFanOut(self,gate):
        self.fanouts.append(gate)
        return len(self.fanouts)-1

    def addFanIn(self,gate):
        self.fanins.append(gate)
        return len(self.fanins)-1

    def addStructureEqualGate(self,gate):
        self.structure_equal_gates.append(gate)

    def getStructureEqualGates(self):
        return self.structure_equal_gates

    def upFlag(self):
        self.flag = True

    def downFlag(self):
        self.flag = False

    def isFlagUp(self):
        return self.flag

    def getTypeStr(self):
        return self.type_str

    def setTypeStr(self,type_str):
        self.type_str = type_str

    def getCopy(self,id):
        gate = copy.copy(self)
        gate.id = id
        gate.fanins = []
        gate.fanouts = []
        return gate

    def relateToPo(self):
        self.ispo = True

    def isPoGate(self):
        return self.ispo

    def getFanOutTypes(self):
        return set([type(fanout) for fanout in self.getFanOuts()])

    def reset(self):
        self.value = None

    def canReason(self):
        if not self.value==None:
            try:
                if self.value==(self.ctrl_value ^ self.phase):
                    return True
            except TypeError:
                pass
            finally:
                assigned_fi = [fi for fi in self.fanins if not fi.value==None]
                if len(assigned_fi)==self.getFanInNum()-1 and all([not fi.value==self.ctrl_value for fi in assigned_fi]):
                    return True
        return False

    def canPropagate(self):
        if all([not fi.value==None for fi in self.fanins]):
            return True
        elif any([fi.value==self.ctrl_value for fi in self.fanins if not fi.value==None]):
            return True 
        else:
            return False

    def reason(self):
        print 'hello'
        try:
            if self.value==(self.ctrl_value^self.phase):
                for fi in self.fanins:
                    #print '------- deduce',fi
                    if not fi.deduceValue(not self.ctrl_value):
                        return False
                return True
        except TypeError:
            one_num = 0
            is_odd = not one_num%2==0
            for fi in self.fanins:
                if fi.value==True:
                    one_num += 1
            for fi in self.fanins:
                if fi.value==None:
                    if self.value==True:
                        #print '------ deduce', fi
                        if not fi.deduceValue(self.phase^is_odd):
                            return False
                    else:
                        #print '------ deduce', fi
                        if not fi.deduceValue(not self.phase^is_odd):
                            return False
            return True
        finally:
            for fi in self.fanins:
                if fi.value==None:
                    #print '------ deduce', fi
                    if not fi.deduceValue(self.ctrl_value):
                        return False
            return True

    def getEncodeConstraint(self):
        added_clas = []
        constraint_type = {'and-not':0,'and-or':0,'and-nor':0,'and-and':0,'or-not':0}
        if self.getFanOutNum() > 1:
            for fanout in self.getFanOuts():
                if isinstance(fanout,AndGate):
                    #print self.getFanOuts()
                    for fanout2 in self.getFanOuts():
                        if isinstance(fanout2,NotGate):
                            continue
                            constraint_type['and-not'] += 1
                            added_clas.append([-fanout.getId(True),-fanout2.getId(True)])
                            added_clas.append([-fanout2.getId(True),-fanout.getId(True)])
                        if isinstance(fanout2,OrGate):
                            constraint_type['and-or'] += 1
                            added_clas.append([-fanout.getId(True),fanout2.getId(True)])
                            added_clas.append([fanout2.getId(True),-fanout.getId(True)])
                        if isinstance(fanout2,NorGate):
                            constraint_type['and-nor'] += 1
                            added_clas.append([-fanout.getId(True),-fanout2.getId(True)])
                            added_clas.append([-fanout2.getId(True),-fanout.getId(True)])
                        if isinstance(fanout2,AndGate):
                            continue
                            constraint_type['and-and'] += 1
                            cla = []
                            for fi in fanout2.getFanIns():
                                if fi==self:
                                    cla.append(-fanout.getId(True))
                                else:
                                    cla.append(-fi.getId(True))
                            cla.append(fanout2.getId(True))
                            added_clas.append(cla)
                        

                        #if isinstance(fanout2,XorGate):
                        #    print 'add and-xor constraint'

                if isinstance(fanout,OrGate):
                    for fanout2 in self.getFanOuts():
                        if isinstance(fanout2,NotGate):
                            constraint_type['or-not'] += 1
                            added_clas.append([fanout.getId(True),fanout2.getId(True)])
                            added_clas.append([fanout2.getId(True),fanout.getId(True)])

        else:
            #print self,'has single fanout'
            pass
        return added_clas, constraint_type

    def idTranslate(self,num):
        self.id += num

    def nameTranslate(self,append_str):
        self.name += append_str
        
class AndGate(Gate):

    def __init__(self,id):
        super(AndGate, self).__init__(id)
        self.type_str = 'AND'
        self.ctrl_value = False
        self.phase = True

    def getProbability(self):
        if not self.probability==None:
            return self.probability
        one_probability = 1
        for fanin in self.fanins:
            one_probability *= fanin.getProbability()
        self.probability = one_probability
        return self.probability

    def eval(self):
        if any([fanin.getValue()==False for fanin in self.fanins]):
            self.value = False
            return self.value
        if any([fanin.getValue()==None for fanin in self.fanins]):
            self.value = None
            return self.value
        self.value = True
        return self.value

    def getSOP(self):
        if not self.SOP==None:
            return self.SOP
        sop = self.fanins[0].getSOP()
        for fanin in self.fanins[1:]:
            sop *= fanin.getSOP()
        self.SOP = sop.clear()
        print self.SOP
        return self.SOP

    def encTseitin(self):
        left_cla = []
        left_cla.append(self.getId(True))
        right_clas = []
        for fanin in self.getFanIns():
            left_cla.append(-fanin.getId(True))
            right_clas.append([-self.getId(True),fanin.getId(True)])
        right_clas.append(left_cla)
        return right_clas

class OrGate(Gate):

    def __init__(self,id):
        super(OrGate, self).__init__(id)
        self.type_str = 'OR'
        self.ctrl_value = True
        self.phase = True

    def getProbability(self):
        if not self.probability==None:
            return self.probability
        zero_probability = 1
        for fanin in self.fanins:
            zero_probability *= (1-fanin.getProbability())
        self.probability = 1 - zero_probability
        return self.probability

    def eval(self):
        if any([fanin.getValue()==True for fanin in self.fanins]):
            self.value = True
            return self.value
        if any([fanin.getValue()==None for fanin in self.fanins]):
            self.value = None
            return self.value
        self.value = False
        return self.value

    def getSOP(self):
        if not self.SOP==None:
            return self.SOP
        sop = self.fanins[0].getSOP()
        for fanin in self.fanins[1:]:
            sop += fanin.getSOP()
        self.SOP = sop.clear()
        print self.SOP
        return self.SOP

    def encTseitin(self):
        right_cla = []
        right_cla.append(-self.getId(True))
        left_clas = []
        for fanin in self.getFanIns():
            right_cla.append(fanin.getId(True))
            left_clas.append([self.getId(True),-fanin.getId(True)])
        left_clas.append(right_cla)
        return left_clas

class XorGate(Gate):

    def __init__(self,id):
        super(XorGate, self).__init__(id)
        self.type_str = 'XOR'
        self.ctrl_value = None
        self.phase = True

    def getProbability(self):
        if not self.probability==None:
            return self.probability
        probability = 0
        for one_num in range(self.getFanInNum()+1):
            if not (one_num % 2 == 0):
                combs = itertools.combinations(range(self.getFanInNum()),one_num)
                for comb in combs:
                    sub_probability = 1
                    for index, fanin in enumerate(self.fanins):
                        if index in comb:
                            sub_probability *= fanin.getProbability()
                        else:
                            sub_probability *= 1-fanin.getProbability()
                    probability += sub_probability
        self.probability = probability
        return self.probability

    def eval(self):
        if any([fanin.getValue()==None for fanin in self.fanins]):
            self.value = None
            return self.value
        else:
            self.value = self.fanins[0].getValue()
            for fanin in self.fanins[1:]:
                self.value = not (self.value==fanin.getValue())
            return self.value

    def getSOP(self):
        if not self.SOP==None:
            return self.SOP
        sop = self.fanins[0].getSOP()
        for fanin in self.fanins[1:]:
            sop = (~sop)*(fanin.getSOP())+(sop)*(~fanin.getSOP())
        self.SOP = sop.clear()
        print self.SOP
        return self.SOP

    def encTseitin(self):
        clas = []
        for one_num in range(self.getFanInNum()+1):
            #if not (one_num % 2 == 0):
            combs = itertools.combinations(range(self.getFanInNum()),one_num)
            for comb in combs:
                cla = []
                for index, fanin in enumerate(self.fanins):
                    if index in comb:
                        cla.append(fanin.getId(True))
                    else:
                        cla.append(-fanin.getId(True))
                if one_num % 2 == 0:
                    cla.append(-self.getId(True))
                else:
                    cla.append(self.getId(True))
                clas.append(cla)
        return clas
            
class BufGate(Gate):

    def __init__(self,id):
        super(BufGate, self).__init__(id)
        self.type_str = 'BUF'
        self.ctrl_value = None
        self.phase = True

    def getProbability(self):
        if not self.probability==None:
            return self.probability
        self.probability = self.fanins[0].getProbability()
        return self.probability

    def eval(self):
        self.value = self.fanins[0].getValue()
        return self.value

    def getSOP(self):
        if not self.SOP==None:
            return self.SOP
        self.SOP = self.fanins[0].getSOP()
        print self.SOP
        return self.SOP

    def encTseitin(self):
        cla1 = [-self.getId(True),self.getFanIn(0).getId(True)] 
        cla2 = [self.getId(True),-self.getFanIn(0).getId(True)]
        return [cla1,cla2]

class NotGate(Gate):

    def __init__(self,id):
        super(NotGate, self).__init__(id)
        self.type_str = 'NOT'
        self.ctrl_value = None
        self.phase = False

    def getProbability(self):
        if not self.probability==None:
            return self.probability
        self.probability = 1 - self.fanins[0].getProbability()
        return self.probability

    def eval(self):
        self.value = not self.fanins[0].getValue()
        return self.value

    def getSOP(self):
        if not self.SOP==None:
            return self.SOP
        self.SOP = ~self.fanins[0].getSOP()
        self.SOP = self.SOP.clear()
        print self.SOP
        return self.SOP

    def encTseitin(self):
        cla1 = [self.getId(True),self.getFanIn(0).getId(True)]
        cla2 = [-self.getId(True),-self.getFanIn(0).getId(True)]
        return [cla1,cla2]

class NorGate(Gate):

    def __init__(self,id):
        super(NorGate, self).__init__(id)
        self.type_str = 'NOR'
        self.ctrl_value = True
        self.phase = False

    def getProbability(self):
        if not self.probability==None:
            return self.probability
        probability = 1
        for fanin in self.fanins:
            probability *= (1-fanin.getProbability())
        self.probability = probability
        return self.probability

    def eval(self):
        if any([fanin.getValue()==True for fanin in self.fanins]):
            self.value = False
            return self.value
        if any([fanin.getValue()==None for fanin in self.fanins]):
            self.value = None
            return self.value
        self.value = True
        return self.value

    def getSOP(self):
        if not self.SOP==None:
            return self.SOP
        sop = self.fanins[0].getSOP()
        for fanin in self.fanins[1:]:
            sop += fanin.getSOP()
        self.SOP = ~sop
        self.SOP = self.SOP.clear()
        print self.SOP
        return self.SOP

    def encTseitin(self):
        left_cla = []
        left_cla.append(self.getId(True))
        right_clas = []
        for fanin in self.getFanIns():
            left_cla.append(fanin.getId(True))
            right_clas.append([-self.getId(True),-fanin.getId(True)])
        right_clas.append(left_cla)
        return right_clas

class NandGate(Gate):

    def __init__(self,id):
        super(NandGate, self).__init__(id)
        self.type_str = 'NAND'
        self.ctrl_value = False
        self.phase = False

    def getProbability(self):
        if not self.probability==None:
            return self.probability
        probability = 1
        for fanin in self.fanins:
            probability *= fanin.getProbability()
        self.probability = 1 - probability
        return self.probability

    def eval(self):
        if any([fanin.getValue()==False for fanin in self.fanins]):
            self.value = True
            return self.value
        if any([fanin.getValue()==None for fanin in self.fanins]):
            self.value = None
            return self.value
        self.value = False
        return self.value

    def getSOP(self):
        if not self.SOP==None:
            return self.SOP
        sop = self.fanins[0].getSOP()
        for fanin in self.fanins[1:]:
            sop *= fanin.getSOP()
        self.SOP = ~sop
        self.SOP = self.SOP.clear()
        print self.SOP
        return self.SOP

    def encTseitin(self):
        right_cla = []
        right_cla.append(-self.getId(True))
        left_clas = []
        for fanin in self.getFanIns():
            right_cla.append(-fanin.getId(True))
            left_clas.append([self.getId(True),fanin.getId(True)])
        left_clas.append(right_cla)
        return left_clas

class XnorGate(Gate):

    def __init__(self,id):
        super(XnorGate, self).__init__(id)
        self.type_str = 'XNOR'
        self.ctrl_value = None
        self.phase = False

    def getProbability(self):
        if not self.probability==None:
            return self.probability
        probability = 0
        for one_num in range(self.getFanInNum()+1):
            if not (one_num % 2 == 0):
                combs = itertools.combinations(range(self.getFanInNum()),one_num)
                for comb in combs:
                    sub_probability = 1
                    for index, fanin in enumerate(self.fanins):
                        if index in comb:
                            sub_probability *= fanin.getProbability()
                        else:
                            sub_probability *= 1-fanin.getProbability()
                    probability += sub_probability
        self.probability = 1-probability
        return self.probability

    def eval(self):
        if any([fanin.getValue()==None for fanin in self.fanins]):
            self.value = None
            return self.value
        else:
            self.value = self.fanins[0].getValue()
            for fanin in self.fanins[1:]:
                self.value = (self.value==fanin.getValue())
            return self.value

    def getSOP(self):
        if not self.SOP==None:
            return self.SOP
        sop = self.fanins[0].getSOP()
        for fanin in self.fanins[1:]:
            sop = (~sop)*(fanin.getSOP())+(sop)*(~fanin.getSOP())
        self.SOP  = ~sop
        self.SOP = self.SOP.clear()
        print self.SOP
        return self.SOP

    def encTseitin(self):
        clas = []
        for one_num in range(self.getFanInNum()+1):
            #if not (one_num % 2 == 0):
            combs = itertools.combinations(range(self.getFanInNum()),one_num)
            for comb in combs:
                cla = []
                for index, fanin in enumerate(self.fanins):
                    if index in comb:
                        cla.append(fanin.getId(True))
                    else:
                        cla.append(-fanin.getId(True))
                if one_num % 2 == 0:
                    cla.append(self.getId(True))
                else:
                    cla.append(-self.getId(True))
                clas.append(cla)
        return clas

class PIGate(Gate):

    def __init__(self,id):
        super(PIGate, self).__init__(id)
        self.type_str = 'PI'
        self.depth = 0
        self.path_num = 1
        self.ctrl_value = None
        self.phase = True

    def getProbability(self):
        return 0.5

    def eval(self):
        self.value = self.value
        return self.value

    def getSOP(self):
        return SOP([Product([self.getId(True)])])

class POGate(Gate):

    def __init__(self,id):
        super(POGate, self).__init__(id)
        self.type_str = 'PO'
        self.ispo = True

class ConstZeroGate(Gate):

    def __init__(self,id):
        super(ConstZeroGate, self).__init__(id)
        self.type_str = 'ConstZero'
        self.value = False
        self.depth = 0
        self.path_num = 1

    def getProbability(self):
        return 0.0

    def eval(self):
        self.value = False
        return self.value

    def getSOP(self):
        return SOP()

    def encTseitin(self):
        cla = [-self.getId(True)]
        return [cla]

class ConstOneGate(Gate):

    def __init__(self,id):
        super(ConstOneGate, self).__init__(id)
        self.type_str = 'ConstOne'
        self.value = True
        self.depth = 0
        self.path_num = 1

    def getProbability(self):
        return 1.0

    def eval(self):
        self.value = True
        return self.value

    def getSOP(self):
        return SOP([Product()])

    def encTseitin(self):
        cla = [self.getId(True)]
        return [cla]



import copy

class SOP(list):
	
	def __init__(self,arg):
		self.parse(arg)

	def parse(self,exp):
		products = exp.split('+')
		self.extend([product.split('*') for product in products])
	
	def __repr__(self):
		return '+'.join([''.join(p) for p in self])
	
	def __str__(self):
		return '+'.join([''.join(p) for p in self])
	
	def __add__(self,other):
		self_cp = copy.deepcopy(self)
		self_cp.extend(other)
		return self_cp

	def __mul__(self,other):
		result = SOP('')
		result.pop()
		if len(other)==1:
			for product in self:
				tmp = product[:]
				tmp.append(other[0])
				result.append(tmp)
		else:
			for product in other:
				result += self * product
		return result

e1 = SOP('a+b')
e2 = SOP('c+d')
#print a




# gate-level parser

def isComment(line):
    if line[0:2]=='//':
        return True
    else:
        return False

def readStatement(reader):
    statement = ''
    while True:
        line = reader.readline()
        if not line:
            return False
        line = line.strip()
        if not line:
            continue
        if isComment(line):
            continue
        if line[-1] == ';' or line=='endmodule':
            return (statement+' '+line).strip()
        else:
            statement += ' '+line

def getCktInfo(file_name):

    ckt_info = {}

    const0 = False
    const1 = False
    verilog_module = ['module','endmodule']
    verilog_pin = {'input':[],'output':[],'wire':[]}
    verilog_gate = {'and':{},'or':{},'xor':{},'nor':{},'nand':{},'xnor':{},'buf':{},'not':{},'zero':{},'one':{}}

    with open(file_name,'r') as reader:
        while True:
            statement = readStatement(reader)
            if not statement:
                break
            tokens = statement.strip(';').split()
            tokens = tokens if len(tokens)>1 or tokens[0]=='endmodule' else statement.strip(';').split('(')
            keyword = tokens.pop(0)
            tokens = ' '.join(tokens)

            keyword = keyword.strip('1234567890')
            if keyword == 'inv':
                keyword = 'not'
            #print keyword,

            if keyword=='module':
                circuit_name = '{0}({1})'.format(tokens.split()[0],file_name)

            elif keyword in verilog_pin:
                tokens = tokens.split()
                pin_names = [token.strip(',') for token in tokens]
                #print pin_names
                verilog_pin[keyword].extend(pin_names)

            elif keyword in verilog_gate:
            	tokens = tokens.partition('(')[2]
            	tokens = tokens.split(',')
            	tokens = [token.strip(' .') for token in tokens]

            	fanins = []

            	for token in tokens:
            		gate_pin_name = token.partition('(')[0]
            		pin_name = token.partition('(')[2]
            		if gate_pin_name=='O':
            			output = pin_name.strip(')')
            		else:
            			fanins.append(pin_name.strip(')'))

                #print output, fanins

                #gate_pins = tokens[1].strip('()').split(',')
                #gate_pins = [pin.strip() for pin in gate_pins]
                #output = gate_pins[0]
                #fanins = gate_pins[1:]
                fanins = ['const_zero' if fanin=="1'b0" else fanin for fanin in fanins]
                fanins = ['const_one' if fanin=="1'b1" else fanin for fanin in fanins]

                #print fanins

                if 'const_zero' in fanins: const0 = True
                if 'const_one' in fanins: const1 = True

                if keyword=='zero':
                    verilog_gate['buf'][output] = ['const_zero']
                    const0 = True
                    continue
                if keyword=='one':
                    verilog_gate['buf'][output] = ['const_one']
                    const1 = True
                    continue

                verilog_gate[keyword][output] = fanins 
            elif keyword in verilog_module:
                pass
            else:
                print 'wrong,wrong,wrong!'

    ckt_info['ckt_name'] = circuit_name
    ckt_info['output'] = verilog_pin['output']
    ckt_info['input'] = verilog_pin['input']
    ckt_info['wire'] = verilog_pin['wire']
    ckt_info['and'] = verilog_gate['and']
    ckt_info['or'] = verilog_gate['or']
    ckt_info['xor'] = verilog_gate['xor']
    ckt_info['nor'] = verilog_gate['nor']
    ckt_info['nand'] = verilog_gate['nand']
    ckt_info['xnor'] = verilog_gate['xnor']
    ckt_info['buf'] = verilog_gate['buf']
    ckt_info['not'] = verilog_gate['not']
    ckt_info['const0'] = const0
    ckt_info['const1'] = const1

    return ckt_info


import copy
import matplotlib.pyplot as plt
import networkx as nx
from gate import *

class CircuitConstructer(object):

    def __init__(self,circuit,ckt_info):

        self.circuit = circuit

        # followings are the must-give info
        self.circuit.circuit_name = ckt_info['ckt_name']
        self.outputs = ckt_info['output']
        self.inputs = ckt_info['input']
        self.wires = ckt_info['wire']
        self.ands = ckt_info['and']
        self.ors = ckt_info['or']
        self.xors = ckt_info['xor']
        self.nors = ckt_info['nor']
        self.nands = ckt_info['nand']
        self.xnors = ckt_info['xnor']
        self.bufs = ckt_info['buf']
        self.nots = ckt_info['not']
        self.const0 = ckt_info['const0']
        self.const1 = ckt_info['const1']

    # main parse function
    def construct(self):
        self.createGates()
        self.connectGates()
        self.relatePO2RealGate()

    # sub parse function
    def createGates(self):
        
        for name in self.inputs:
            gate = PIGate(len(self.circuit.gates))
            gate.setName(name)
            self.circuit.addGate(gate)

        for name in self.outputs:
            gate = POGate(len(self.circuit.gates))
            gate.setName(name)
            self.circuit.addGate(gate)

        self.circuit.po_end_index = len(self.circuit.gates)
            
        for name in self.ands:
            gate = AndGate(len(self.circuit.gates))
            gate.setName(name)
            self.circuit.addGate(gate)
            
        for name in self.ors:
            gate = OrGate(len(self.circuit.gates))
            gate.setName(name)
            self.circuit.addGate(gate)
            
        for name in self.xors:
            gate = XorGate(len(self.circuit.gates))
            gate.setName(name)
            self.circuit.addGate(gate)
            
        for name in self.nors:
            gate = NorGate(len(self.circuit.gates))
            gate.setName(name)
            self.circuit.addGate(gate)
            
        for name in self.nands:
            gate = NandGate(len(self.circuit.gates))
            gate.setName(name)
            self.circuit.addGate(gate)

        for name in self.xnors:
            gate = XnorGate(len(self.circuit.gates))
            gate.setName(name)
            self.circuit.addGate(gate)
        
        for name in self.bufs:
            gate = BufGate(len(self.circuit.gates))
            gate.setName(name)
            self.circuit.addGate(gate)
            
        for name in self.nots:
            gate = NotGate(len(self.circuit.gates))
            gate.setName(name)
            self.circuit.addGate(gate)

        if self.const0:
            gate = ConstZeroGate(len(self.circuit.gates))
            gate.setName('const_zero')
            self.circuit.addGate(gate)

        if self.const1:
            gate = ConstOneGate(len(self.circuit.gates))
            gate.setName('const_one')
            self.circuit.addGate(gate)

    def connectGates(self):
        for name in self.ands:
            gate = self.circuit.getGateFromName(name)
            for fanin_name in self.ands[name]:
                fanin_gate = self.circuit.getGateFromName(fanin_name)
                gate.addFanIn(fanin_gate)
                fanin_gate.addFanOut(gate)
        for name in self.ors:
            gate = self.circuit.getGateFromName(name)
            for fanin_name in self.ors[name]:
                fanin_gate = self.circuit.getGateFromName(fanin_name)
                gate.addFanIn(fanin_gate)
                fanin_gate.addFanOut(gate)
        for name in self.xors:
            gate = self.circuit.getGateFromName(name)
            for fanin_name in self.xors[name]:
                fanin_gate = self.circuit.getGateFromName(fanin_name)
                gate.addFanIn(fanin_gate)
                fanin_gate.addFanOut(gate)
        for name in self.nors:
            gate = self.circuit.getGateFromName(name)
            for fanin_name in self.nors[name]:
                fanin_gate = self.circuit.getGateFromName(fanin_name)
                gate.addFanIn(fanin_gate)
                fanin_gate.addFanOut(gate)
        for name in self.nands:
            gate = self.circuit.getGateFromName(name)
            for fanin_name in self.nands[name]:
                fanin_gate = self.circuit.getGateFromName(fanin_name)
                gate.addFanIn(fanin_gate)
                fanin_gate.addFanOut(gate)
        for name in self.xnors:
            gate = self.circuit.getGateFromName(name)
            for fanin_name in self.xnors[name]:
                fanin_gate = self.circuit.getGateFromName(fanin_name)
                gate.addFanIn(fanin_gate)
                fanin_gate.addFanOut(gate)
        for name in self.bufs:
            gate = self.circuit.getGateFromName(name)
            for fanin_name in self.bufs[name]:
                fanin_gate = self.circuit.getGateFromName(fanin_name)
                gate.addFanIn(fanin_gate)
                fanin_gate.addFanOut(gate)
        for name in self.nots:
            gate = self.circuit.getGateFromName(name)
            for fanin_name in self.nots[name]:
                fanin_gate = self.circuit.getGateFromName(fanin_name)
                gate.addFanIn(fanin_gate)
                fanin_gate.addFanOut(gate)

    def relatePO2RealGate(self):

        for po_gate in self.circuit.getPOGates():
            po_id = po_gate.getId()
            for gate in self.circuit.gates[self.circuit.po_end_index:]:
                if gate.getName()==po_gate.getName():
                    gate.setTypeStr('PO-'+gate.getTypeStr())
                    self.circuit.gates[po_id] = gate
                    self.circuit.gates[po_id].relateToPo()


class Circuit:

    def __init__(self):

        self.circuit_name = None
        self.gates = []

        self.name2id_dic = {}
        self.id2name_dic = {}
        self.po_end_index = None

        self.num_pi = 0
        self.num_po = 0
        self.num_and = 0
        self.num_or = 0
        self.num_xor = 0
        self.num_nor = 0
        self.num_nand = 0
        self.num_xnor = 0
        self.num_buf = 0
        self.num_not = 0
        self.num_const0 = 0
        self.num_const1 = 0

        self.topological_orderd_gates = None

    def __str__(self):
        return '{0}'.format(self.circuit_name)

    # input
    def constructCkt(self,ckt_info):
        ckt_constructer = CircuitConstructer(self,ckt_info)
        ckt_constructer.construct()
        self.computeGateNum()

    def computeGateNum(self):
        for index, gate in enumerate(self.gates):
            if index < self.po_end_index:
                if isinstance(gate,PIGate):
                    self.num_pi += 1
                else:
                    self.num_po += 1
            else:
                if isinstance(gate,AndGate):
                    self.num_and += 1
                elif isinstance(gate,OrGate):
                    self.num_or += 1
                elif isinstance(gate,XorGate):
                    self.num_xor += 1
                elif isinstance(gate,NorGate):
                    self.num_nor += 1
                elif isinstance(gate,NandGate):
                    self.num_nand += 1
                elif isinstance(gate,XnorGate):
                    self.num_xnor += 1
                elif isinstance(gate,BufGate):
                    self.num_buf += 1
                elif isinstance(gate,NotGate):
                    self.num_not += 1
                elif isinstance(gate,ConstZeroGate):
                    self.num_const0 += 1
                elif isinstance(gate,ConstOneGate):
                    self.num_const1 += 1

    # setter
    def setName(self,name):
        self.circuit_name = name

    def idTranslate(self,num):
        for gate in self.getGatesOnce():
            gate.idTranslate(num)

    def gateNamesTranslate(self,append_str):
        for gate in self.getGatesOnce():
            gate.nameTranslate(append_str)

    # getter(gate num)
    def getPiNum(self):
        return self.num_pi

    def getPoNum(self):
        return self.num_po

    def getAndNum(self):
        return self.num_and

    def getOrNum(self):
        return self.num_or

    def getXorNum(self):
        return self.num_xor

    def getNorNum(self):
        return self.num_nor

    def getNandNum(self):
        return self.num_nand

    def getXnorNum(self):
        return self.num_xnor

    def getBufNum(self):
        return self.num_buf

    def getNotNum(self):
        return self.num_not

    def getConstZeroNum(self):
        return self.num_const0

    def getConstOneNum(self):
        return self.num_const1

    def getGateCount(self):
        gate_count = 0
        for gate in self.gates[:self.po_end_index]:
            if isinstance(gate,PIGate):
                gate_count += 1
        gate_count += len(self.gates[self.po_end_index:])
        return gate_count

    # getter(circuit properties)
    def getAllPOProbability(self):
        return [PO.getProbability() for PO in self.getPOGates()]

    def getPOProbability(self,index):
        return self.getPOGates()[index].getProbability()

    def getPhaseConsistentRate(self):
        num_of_not_consistent = 0
        for gate in self.gates:
            for index, fanout in enumerate(gate.getFanOuts()):
                if index==0:
                    consistent = isinstance(fanout,NotGate)
                else:
                    if not consistent==isinstance(fanout,NotGate):
                        num_of_not_consistent += 1
                        break
        return float(num_of_not_consistent)/float(self.getGateCount()) 

    def getAvgFanOutNum(self):
        total_fanout_num = 0
        for gate in self.gates[:self.po_end_index]:
            if isinstance(gate,PIGate):
                total_fanout_num += gate.getFanOutNum()
        for gate in self.gates[self.po_end_index:]:
            total_fanout_num += gate.getFanOutNum()
        return float(total_fanout_num)/float(self.getGateCount())
    
    def getAvgFanInNum(self):
        total_fanin_num = 0
        for gate in self.gates[self.po_end_index:]:
            total_fanin_num += gate.getFanInNum()
        return float(total_fanin_num)/float(self.getGateCount())

    def getMaxFanOutNum(self):
        max_fanout_num = 0
        for gate in self.gates[:self.po_end_index]:
            if isinstance(gate,PIGate):
                if gate.getFanOutNum() > max_fanout_num:
                    max_fanout_num = gate.getFanOutNum()
        for gate in self.gates[self.po_end_index:]:
            if gate.getFanOutNum() > max_fanout_num:
                max_fanout_num = gate.getFanOutNum()
        return max_fanout_num

    def getMaxFanInNum(self):
        max_fanin_num = 0
        for gate in self.gates[:self.po_end_index]:
            if isinstance(gate,PIGate):
                if gate.getFanInNum() > max_fanin_num:
                    max_fanin_num = gate.getFanInNum()
        for gate in self.gates[self.po_end_index:]:
            if gate.getFanInNum() > max_fanin_num:
                max_fanin_num = gate.getFanInNum()
        return max_fanin_num

    def getFanOutNumVariance(self):
        avg_fanout_num = self.getAvgFanOutNum()
        gates = [gate for gate in self.gates[:self.po_end_index] if isinstance(gate,PIGate)]
        gates.extend(self.gates[self.po_end_index:])
        sigma_result = 0
        for gate in gates:
            sigma_result += (gate.getFanOutNum()-avg_fanout_num)**2
        return float(sigma_result)/float(self.getGateCount())

    def getFanInNumVariance(self):
        avg_fanin_num = self.getAvgFanInNum()
        gates = [gate for gate in self.gates[:self.po_end_index] if isinstance(gate,PIGate)]
        gates.extend(self.gates[self.po_end_index:])
        sigma_result = 0
        for gate in gates:
            sigma_result += (gate.getFanInNum()-avg_fanin_num)**2
        return float(sigma_result)/float(self.getGateCount())

    def getMaxFanInDepthDiff(self):
        max_fanin_depth_diff = 0
        for gate in self.gates[self.po_end_index:]:
            if isinstance(gate,ConstOneGate) or isinstance(gate,ConstZeroGate):
                continue
            max_fanin_depth = max([fanin.getDepth() for fanin in gate.getFanIns()])
            min_fanin_depth = min([fanin.getDepth() for fanin in gate.getFanIns()])
            depth_diff = max_fanin_depth - min_fanin_depth
            if depth_diff > max_fanin_depth_diff:
                max_fanin_depth_diff = depth_diff
        return max_fanin_depth_diff

    def getAvgFanInDepthDiff(self):
        total_fanin_depth_diff = 0
        for gate in self.gates[self.po_end_index:]:
            if isinstance(gate,ConstOneGate) or isinstance(gate,ConstZeroGate):
                continue
            max_fanin_depth = max([fanin.getDepth() for fanin in gate.getFanIns()])
            min_fanin_depth = min([fanin.getDepth() for fanin in gate.getFanIns()])
            depth_diff = max_fanin_depth - min_fanin_depth
            total_fanin_depth_diff += depth_diff
        return float(total_fanin_depth_diff)/float(self.getGateCount())

    # getter(gate/id)
    def getGates(self):
        return self.gates

    def getGatesOnce(self):
        return self.getPIGates() + self.gates[self.po_end_index:]

    def getPIGates(self):
        return [gate for gate in self.gates[:self.po_end_index] if isinstance(gate,PIGate)]

    def getAndGates(self):
        return [gate for gate in self.gates[self.po_end_index:] if isinstance(gate,AndGate)]

    def getOrGates(self):
        return [gate for gate in self.gates[self.po_end_index:] if isinstance(gate,OrGate)]

    def getXorGates(self):
        return [gate for gate in self.gates[self.po_end_index:] if isinstance(gate,XorGate)]

    def getNandGates(self):
        return [gate for gate in self.gates[self.po_end_index:] if isinstance(gate,NandGate)]

    def getNorGates(self):
        return [gate for gate in self.gates[self.po_end_index:] if isinstance(gate,NorGate)]

    def getXnorGates(self):
        return [gate for gate in self.gates[self.po_end_index:] if isinstance(gate,XnorGate)]

    def getBufGates(self):
        return [gate for gate in self.gates[self.po_end_index:] if isinstance(gate,BufGate)]

    def getNotGates(self):
        return [gate for gate in self.gates[self.po_end_index:] if isinstance(gate,NotGate)]

    def getConstZeroGates(self):
        return [gate for gate in self.gates[self.po_end_index:] if isinstance(gate,ConstZeroGate)]

    def getConstOneGates(self):
        return [gate for gate in self.gates[self.po_end_index:] if isinstance(gate,ConstOneGate)]

    def getIdFromName(self,name):
        return self.name2id_dic[name]

    def getGateFromId(self,id):
        return self.gates[id]

    def getGateFromName(self,name):
        return self.getGateFromId(self.name2id_dic[name])

    def getPOGates(self):
        po_gates = []
        for gate in self.gates[:self.po_end_index]:
            if gate.isPoGate():
                po_gates.append(gate)
        return po_gates

    def getGatesInTopologicalOrder(self,reverse=False):
        if self.topological_orderd_gates:
            return self.topological_orderd_gates
        else:
            self.downGateFlag()
            gates = []
            tmp_gates = self.gates[:self.getPiNum()]+self.getConstOneGates()[:]+self.getConstZeroGates()[:]
            #if self.getConstZeroNum==1:
            #    tmp_gates.append(self.gates[-2])
            #if self.getConstOneNum==1:
            #    tmp_gates.append(self.gates[-1])
            for gate in tmp_gates:
                gate.upFlag()
            while tmp_gates:
                gate = tmp_gates.pop(0)
                for fanout in gate.getFanOuts():
                    if fanout.isFlagUp():
                        continue
                    if all([fanin.isFlagUp() for fanin in fanout.getFanIns()]):
                        tmp_gates.append(fanout)
                        fanout.upFlag()
                gates.append(gate)
            self.topological_orderd_gates = gates if not reverse else gates.reverse()
            return self.topological_orderd_gates

    # construct
    def addGate(self,gate):
        self.gates.append(gate)
        self.registerGate(gate)

    def registerGate(self,gate):
        self.name2id_dic[gate.getName()] = gate.getId()
        self.id2name_dic[gate.getId()] = gate.getName()

    # evaluate
    def evaluateCkt(self):
        for gate in self.getGatesInTopologicalOrder():
            gate.eval()

    def setPIValueZero(self):
        for pi in self.gates[:self.po_end_index]:
            if isinstance(pi,PIGate):
                pi.setValue(False)

    def setPIValueOne(self):
        for pi in self.gates[:self.po_end_index]:
            if isinstance(pi,PIGate):
                pi.setValue(True)

    def resetAllGates(self):
        for gate in self.gates:
            gate.reset()

    # cone and subckt
    def getConeCircuit(self,gate):
        assert gate in self.gates
        ckt = Circuit()
        self.traverseCone(gate)
        oldid2newid_dic = {}
        newid2oldid_dic = {}

        for id, gate in enumerate(self.gates):
            if gate.isFlagUp():
                oldid2newid_dic[id] = len(ckt.gates)
                newid2oldid_dic[len(ckt.gates)] = id
                copy_gate = gate.getCopy(len(ckt.gates))
                copy_gate.setName(gate.getName())
                ckt.addGate(copy_gate)

        for gate in self.gates[self.po_end_index:]:
            if gate.isFlagUp():
                #print gate, 'setting fanin and fanout ...'
                #print '\t fanins:', gate.getFanIns()
                for fanin in gate.getFanIns():
                    #print '\t fanin:', fanin, 'check flag ...'
                    if fanin.isFlagUp():
                        #print '\t\t fanin:', fanin, 'is flag up !'
                        new_fanin = ckt.getGateFromId(oldid2newid_dic[fanin.getId()])
                        new_gate = ckt.getGateFromId(oldid2newid_dic[gate.getId()])
                        new_gate.addFanIn(new_fanin)
                        new_fanin.addFanOut(new_gate)
    
        ckt.po_end_index = self.po_end_index

        for gate in ckt.gates:
            if gate.isPoGate():
                #print gate, 'is a po gate, its id:', gate.getId(),'its old id:',newid2oldid_dic[gate.getId()], 'cp with:', self.po_end_index
                if newid2oldid_dic[gate.getId()] < self.po_end_index:
                    ckt.po_end_index = gate.getId()

        if ckt.po_end_index==self.po_end_index:
            for gate in ckt.gates:
                if isinstance(gate,PIGate):
                    ckt.po_end_index = gate.getId()

        ckt.po_end_index += 1

        #print 'po end index of cone:', ckt.po_end_index

        for po_gate in ckt.getPOGates():
            po_id = po_gate.getId()
            for gate in ckt.gates[ckt.po_end_index:]:
                if gate.getName()==po_gate.getName():
                    ckt.gates[po_id] = gate

        ckt.computeGateNum()
        
        return ckt

    def traverseCone(self,gate):
        assert gate in self.gates
        self.downGateFlag()
        
        gates = [gate]
        while(gates):
            gate = gates.pop()
            #print 'now go to', gate, 'remain num in gates', len(gates)
            if gate.isFlagUp():
                continue
            gate.upFlag()
            for fanin_gate in gate.getFanIns():
                gates.append(fanin_gate)

    def isConeClose(self,gate):
        assert gate in self.gates
        #print 'cone',gate
        self.traverseCone(gate)
        for g in self.getPIGates()+self.getGates()[self.po_end_index:]:
            if g.isFlagUp() and not g==gate:
                #print '...check', g
                for fo in g.getFanOuts():
                    #print '......',fo
                    if not fo.isFlagUp():
                        #print 'error------------'
                        #print g, 'has outer fanout', fo
                        return False
        return True

    def getAllCloseCone(self):
        gates = []
        for gate in self.gates[self.po_end_index:]:
            if self.isConeClose(gate):
                gates.append(gate)
        return gates

    def downGateFlag(self):
        for gate in self.gates:
            gate.downFlag()

    # encode
    def encTseitin(self,file_name,option='sat'):
        with open(file_name,'with') as writer:
            print >>writer,'c cnf written by miniCkt'
            for gate in self.getGates()[self.po_end_index:]:
                for cla in gate.encTseitin():
                    for lit in cla:
                        print >>writer,lit,
                    print >>writer,'0'
            if option=='sat':
                for gate in self.getPOGates():
                    print >>writer,gate.getId(True),
                    print >>writer,'0'
            elif option=='unsat':
                for gate in self.getPOGates():
                    print >> writer,-gate.getId(True),
                    print >>writer,'0'
            elif option=='pure':
                pass

    def encReason(self,file_name):
        self.encTseitin(file_name)
        with open(file_name,'a') as writer:
            for gate in self.getAndGates()+self.getNorGates():
                self.resetAllGates()
                gate.setValue(True)
                self.reason()
                for g in self.getGatesInTopologicalOrder():
                    if not g.getValue()==None and g not in gate.getFanIns() and g not in gate.getFanOuts():
                        if g.getValue()==True:
                            print >>writer, -gate.getId(True), g.getId(True), '0'
                        else:
                            print >>writer, -gate.getId(True), -g.getId(True), '0'
            for gate in self.getOrGates()+self.getNandGates():    
                self.resetAllGates()
                gate.setValue(False)
                self.reason()
                for g in self.getGatesInTopologicalOrder():
                    if not g.getValue()==None and g not in gate.getFanIns() and g not in gate.getFanOuts():
                        if g.getValue()==True:
                            print >>writer, gate.getId(True), g.getId(True), '0'
                        else:
                            print >>writer, gate.getId(True), -g.getId(True), '0'


    # reason
    def reason(self):
        todo_gates = []
        for gate in self.getPIGates()+self.gates[self.po_end_index:]:
            #if any([fi.getValue()==None for fi in gate.getFanIns()]) or gate.getValue()==None:
            todo_gates.append(gate)
        check_num = len(todo_gates)
        while True:
            todo_gate_index = []
            for index, gate in enumerate(todo_gates):
                if gate.canReason():
                    #print '--- reason',gate
                    if not gate.reason():
                        return False
                elif gate.canPropagate():
                    #print '--- eval',gate
                    ori_value = gate.getValue()
                    if not ori_value==None and not gate.eval()==ori_value:
                        return False
                    else:
                        gate.eval()
                else:
                    todo_gate_index.append(index)
            temp_gates = []
            for index, gate in enumerate(todo_gates):
                if index in todo_gate_index:
                    temp_gates.append(gate)
            todo_gates = temp_gates
            if len(todo_gates)==check_num:
                break
            else:
                check_num = len(todo_gates)
        return True

    # show
    def showCircuitGraph(self):
        G = nx.DiGraph()
        for gate in self.gates:
            for fanout in gate.getFanOuts():
                G.add_edge(gate,fanout)
        nx.draw(G, node_size=1000, node_color='white')
        plt.show()

    def showGatesRatio(self, ):
        labels = ['PI', 'AND', 'OR', 'XOR', 'XNOR', 'NAND', 'NOR', 'NOT', 'BUF', 'Const0', 'Const1']
        sizes = []
        sizes.append(self.getPiNum())
        sizes.append(self.getAndNum())
        sizes.append(self.getOrNum())
        sizes.append(self.getXorNum())
        sizes.append(self.getXnorNum())
        sizes.append(self.getNandNum())
        sizes.append(self.getNorNum())
        sizes.append(self.getNotNum())
        sizes.append(self.getBufNum())
        sizes.append(self.getConstZeroNum())
        sizes.append(self.getConstOneNum())
        for index, size in enumerate(sizes):
            labels[index] = labels[index] + '[' + str(size) +']'
        rm_index = []
        for index, size in enumerate(sizes):
            if size==0 or float(size)/float(self.getGateCount())<0.05:
                rm_index.append(index)
        labels = [labels[index] for index in range(len(sizes)) if index not in rm_index]
        sizes = [sizes[index] for index in range(len(sizes)) if index not in rm_index]
        explode = []
        for i in range(len(sizes)):
            if sizes[i]==max(sizes):
                explode.append(0.1)
            else:
                explode.append(0)

        plt.pie(sizes, labels=labels, explode=explode, autopct='%1.1f%%', shadow=True, startangle=90)
        plt.axis('equal')
        plt.show()

    # test
    def testtopo(self,gates):
        self.downGateFlag()
        for gate in gates:
            for fanin in gate.getFanIns():
                if not fanin.isFlagUp():
                    print 'wrong', gate, '=>', fanin, 'is not traversed' 
                    exit()
            gate.upFlag()

    def printFanOutTypes(self):
        for gate in self.gates:
            print gate.getFanOutTypes()


from distutils.core import setup
setup(
name = 'minickt',
packages = ['minickt', 'cktParser', 'cktWriter'],
scripts = ['minic'],
install_requires=['matplotlib','networkx'],
version = '0.8.4',
description = 'Analyzer and simulator of logic circuit',
author = 'dokelung',
author_email = 'dokelung@gmail.com',
url = 'https://github.com/dokelung/minickt',
download_url = 'https://github.com/dokelung/minickt/tarball/v0.8.4',
keywords = ['circuit','eda','verification','logic','gate'],
classifiers = [],
)



from circuit import *

class CircuitManager(object):

	def __init__(self):
		self.ckts = []
		self.token = 0

	def getToken(self):
		return self.token

	def transToCkt(self,index):
		self.token = index

	def getCurrentCkt(self):
		return self.ckts[self.token]

	def addCkt(self,ckt):
		self.ckts.append(ckt)

	def rmCktByIndex(self,index):
		self.ckts.pop(index)
		if self.token < index:
			pass
		else:
			self.token -= 1

	def getCkts(self):
		return self.ckts

	def getCkt(self,index):
		return self.ckts[index]

	def changeCktName(self,name):
		self.ckts[self.token].setName(name)
